from textwrap import dedent

from clypi import boxed


def test_boxed():
    result = boxed("a" * 16, width=20)
    expected = dedent(
        """
        ┏━━━━━━━━━━━━━━━━━━┓
        ┃ aaaaaaaaaaaaaaaa ┃
        ┗━━━━━━━━━━━━━━━━━━┛
        """
    )
    assert result == expected.strip()


def test_boxed_with_wrapping():
    result = boxed("a" * 40, width=20)
    expected = dedent(
        """
        ┏━━━━━━━━━━━━━━━━━━┓
        ┃ aaaaaaaaaaaaaaaa ┃
        ┃ aaaaaaaaaaaaaaaa ┃
        ┃ aaaaaaaa         ┃
        ┗━━━━━━━━━━━━━━━━━━┛
        """
    )
    assert result == expected.strip()



import asyncio
import re
import shutil
import time
import tomllib
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import anyio
from typing_extensions import override

import clypi.parsers as cp
from clypi import Command, Positional, Spinner, arg, boxed, cprint, style

MDTEST_DIR = Path.cwd() / "mdtest_autogen"
PREAMBLE = """\
from pathlib import Path  # pyright: ignore
from typing import reveal_type, Any  # pyright: ignore
from clypi import *  # pyright: ignore
from enum import Enum  # pyright: ignore
from datetime import datetime, timedelta  # pyright: ignore

def assert_raises(func: Any) -> Any:  
    exc = None
    try:
        func()
    except Exception as e:
        exc = e
    assert exc is not None
"""


class TestFailed(Exception):
    pass


@dataclass
class Test:
    name: str
    orig: str
    code: str
    args: str = ""
    stdin: str = ""

    @property
    def command(self) -> str:
        return ""


@dataclass
class RunTest(Test):
    @property
    @override
    def command(self) -> str:
        cmd = f"uv run --all-extras {MDTEST_DIR / self.name}.py"
        if self.args:
            cmd += f" {self.args}"
        return cmd


@dataclass
class RunPyright(Test):
    @property
    @override
    def command(self) -> str:
        return f"uv run --all-extras pyright {MDTEST_DIR / self.name}.py"


async def parse_file(sm: asyncio.Semaphore, file: Path) -> list[Test]:
    tests: list[Test] = []
    base_name = file.as_posix().replace("/", "-").replace(".md", "").lower()

    # Wait for turn
    await sm.acquire()

    async with await anyio.open_file(file, "r") as f:
        current_test: list[str] = []
        in_test, args, stdin = False, "", ""
        async for line in f:
            # End of a code block
            if "```" in line and current_test:
                code = "\n".join(current_test[1:])
                tests.extend(
                    [
                        RunTest(
                            name=f"{base_name}-{len(tests)}-run",
                            orig=dedent(code),
                            code=PREAMBLE + dedent(code),
                            args=args,
                            stdin=stdin + "\n",
                        ),
                        RunPyright(
                            name=f"{base_name}-{len(tests)}-pyright",
                            orig=dedent(code),
                            code=PREAMBLE + dedent(code),
                        ),
                    ]
                )
                in_test, current_test, args, stdin = False, [], "", ""

            # We're in a test, accumulate all lines
            elif in_test:
                current_test.append(line.removeprefix("> ").removeprefix(">").rstrip())

            # Mdtest arg definition
            elif g := re.search("<!-- mdtest-args (.*) -->", line):
                args = g.group(1)
                in_test = True

            # Mdtest stdin definition
            elif g := re.search("<!-- mdtest-stdin (.*) -->", line):
                stdin = g.group(1)
                in_test = True

            # Mdtest generic definition
            elif g := re.search("<!-- mdtest -->", line):
                in_test = True

            elif "mdtest" in line:
                raise ValueError(f"Invalid mdtest config line: {line}")

    sm.release()
    cprint(style("✔", fg="green") + f" Collected {len(tests)} tests for {file}")
    return tests


def error_msg(test: Test, stdout: str | None = None, stderr: str | None = None) -> str:
    error: list[str] = []
    error.append(style(f"\n\nError running test {test.name!r}\n", fg="red", bold=True))
    error.append(boxed(test.orig, title="Code", width="max"))

    if stdout:
        error.append("")
        error.append(boxed(stdout.strip(), title="Stdout", width="max"))

    if stderr:
        error.append("")
        error.append(boxed(stderr.strip(), title="Stderr", width="max"))

    return "\n".join(error)


class Runner:
    def __init__(self, parallel: int, timeout: int, verbose: bool) -> None:
        self.sm = asyncio.Semaphore(parallel)
        self.timeout = timeout
        self.verbose = verbose

    async def run_test(self, test: Test) -> tuple[str, str | None]:
        # Save test to temp file
        test_file = MDTEST_DIR / f"{test.name}.py"
        test_file.write_text(test.code)

        # Await the subprocess to run it
        proc = await asyncio.create_subprocess_shell(
            test.command,
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        try:
            stdout, stderr = await proc.communicate(test.stdin.encode())
        except:
            proc.terminate()
            raise

        # If no errors, return
        if proc.returncode == 0:
            if self.verbose:
                cprint(
                    style("✔", fg="green")
                    + f" Test {test.name} passed with command: {test.command}"
                )
            return test.name, None

        # If there was an error, pretty print it
        error = error_msg(test, stdout.decode(), stderr.decode())
        return test.name, error

    async def run_test_with_timeout(self, test: Test) -> tuple[str, str | None]:
        await self.sm.acquire()
        start = time.perf_counter()
        try:
            async with asyncio.timeout(self.timeout):
                return await self.run_test(test)
        except TimeoutError:
            error = error_msg(
                test,
                stderr=f"Test timed out after {time.perf_counter() - start:.3f}s",
            )
            return test.name, error
        finally:
            self.sm.release()

    async def run_mdtests(self, tests: list[Test]) -> int:
        errors: list[str] = []
        async with Spinner("Running Markdown Tests", capture=True) as s:
            coros = [self.run_test_with_timeout(test) for test in tests]
            for task in asyncio.as_completed(coros):
                idx, err = await task
                if err is None:
                    cprint(style("✔", fg="green") + f" Finished test {idx}")
                else:
                    errors.append(err)
                    cprint(style("×", fg="red") + f" Finished test {idx}")

            if errors:
                await s.fail()

        for err in errors:
            cprint(err)

        return 1 if errors else 0


class Mdtest(Command):
    """
    Run python code embedded in markdown files to ensure it's
    runnable.
    """

    files: Positional[list[Path] | None] = arg(
        help="The list of markdown files to test",
        default=None,
    )
    parallel: int | None = arg(None, parser=cp.Int(positive=True))
    timeout: int = arg(6, parser=cp.Int(positive=True))
    config: Path = Path("./pyproject.toml")
    verbose: bool = arg(False, help="Enable verbose output", short="v")

    def load_config(self):
        if not self.config.exists():
            return Mdtest()

        with open(self.config, "rb") as f:
            conf = tomllib.load(f)

        with suppress(KeyError):
            data = conf["tool"]["mdtest"]
            parallel = int(data["parallel"]) if "parallel" in data else None
            files = [p for f in data["include"] for p in Path().glob(f)]
            return Mdtest(files, parallel)

        return Mdtest()

    @override
    async def run(self) -> None:
        conf = self.load_config()
        files = self.files or conf.files
        parallel = self.parallel or conf.parallel or 1
        if files is None:
            cprint("No files to run!", fg="yellow")
            return

        # Setup test dir
        MDTEST_DIR.mkdir(exist_ok=True)

        # Assert each file exists
        for file in files:
            assert file.exists(), f"File {file} does not exist!"

        try:
            # Collect tests
            async with Spinner("Collecting Markdown Tests", capture=True):
                sm = asyncio.Semaphore(parallel)
                per_file = await asyncio.gather(
                    *(
                        parse_file(sm, file)
                        for file in files
                        if not file.parents[-1].name.startswith(".")
                    )
                )
                all_tests = [test for file in per_file for test in file]

            # Run each file
            print()
            code = await Runner(parallel, self.timeout, self.verbose).run_mdtests(
                all_tests
            )
        finally:
            # Cleanup
            shutil.rmtree(MDTEST_DIR)

        raise SystemExit(code)


if __name__ == "__main__":
    mdtest = Mdtest.parse()
    mdtest.start()



from __future__ import annotations

import os
from enum import Enum, auto

from clypi._colors import remove_style


class Unset(Enum):
    TOKEN = auto()


UNSET = Unset.TOKEN


def visible_width(s: str) -> int:
    s = remove_style(s)
    return len(s)


def get_term_width():
    if width := os.getenv("CLYPI_TERM_WIDTH"):
        return int(width)

    try:
        return os.get_terminal_size().columns
    except OSError:
        from clypi._configuration import get_config

        return get_config().fallback_term_width


def trim_split_collection(s: str):
    if ", " in s:
        return s.split(", ")
    return s.split(",")



import enum
import typing as t
from datetime import UTC, datetime, timedelta
from pathlib import Path

import pytest

import clypi.parsers as cp
from clypi import Positional


class Color(enum.Enum):
    RED = 1
    BLUE = 2


SUCCESS_PRIMITIVES = [
    ("1", cp.Int(), 1),
    ("1", cp.Float(), 1),
    ("1.2", cp.Float(), 1.2),
    ("y", cp.Bool(), True),
    ("yes", cp.Bool(), True),
    ("tRUe", cp.Bool(), True),
    ("n", cp.Bool(), False),
    ("No", cp.Bool(), False),
    ("faLse", cp.Bool(), False),
    ("", cp.Str(), ""),
    ("a", cp.Str(), "a"),
    ("2025-03-12", cp.DateTime(), datetime(2025, 3, 12)),
    ("2025/3/12", cp.DateTime(), datetime(2025, 3, 12)),
    (
        "2025-03-12T00:00:00Z",
        cp.DateTime(),
        datetime(2025, 3, 12, 0, 0, 0, tzinfo=UTC),
    ),
    ("1d", cp.TimeDelta(), timedelta(days=1)),
    ("1 day", cp.TimeDelta(), timedelta(days=1)),
    ("2weeks", cp.TimeDelta(), timedelta(weeks=2)),
    ("./tests/parsers_test.py", cp.Path(), Path("./tests/parsers_test.py")),
    (
        "./tests/parsers_test.py",
        cp.Path(exists=True),
        Path("./tests/parsers_test.py"),
    ),
    ("y", cp.Union(cp.Int(), cp.Bool()), True),
    ("1", cp.Union(cp.Int(), cp.Bool()), 1),
    ("1", cp.Literal(1, "foo"), 1),
    ("foo", cp.Literal(1, "foo"), "foo"),
    ("red", cp.Enum(Color), Color.RED),
    ("blue", cp.Enum(Color), Color.BLUE),
    ("none", cp.NoneParser(), None),
    ("", cp.NoneParser(), None),
    ("", cp.Str() | cp.NoneParser(), None),
    ("", cp.NoneParser() | cp.Str(), None),
]

FAILURE_PRIMITIVES = [
    ("a", cp.Int()),
    ("1.1", cp.Int()),
    ("a", cp.Float()),
    ("p", cp.Bool()),
    ("falsey", cp.Bool()),
    ("", cp.Bool()),
    ("lsf2", cp.DateTime()),
    ("1 month", cp.TimeDelta()),
    ("1y", cp.TimeDelta()),
    (
        "./tests/parsers_test2.py",
        cp.Path(exists=True),
    ),
    ("a", cp.Union(cp.Int(), cp.Bool())),
    ("2", cp.Literal("1", "foo")),
    ("green", cp.Enum(Color)),
    ("a", cp.NoneParser()),
]


@pytest.mark.parametrize("value,parser,expected", SUCCESS_PRIMITIVES)
def test_successfull_parsers(value: t.Any, parser: cp.Parser[t.Any], expected: t.Any):
    assert parser(value) == expected


@pytest.mark.parametrize("value,parser", FAILURE_PRIMITIVES)
def test_failed_parsers(value: t.Any, parser: cp.Parser[t.Any]):
    with pytest.raises(Exception):
        parser(value)


@pytest.mark.parametrize(
    "value,parser,expected", [(v, cp.List(p), [e]) for (v, p, e) in SUCCESS_PRIMITIVES]
)
def test_successfull_list_parsers(
    value: t.Any, parser: cp.Parser[t.Any], expected: t.Any
):
    assert parser(value) == expected


@pytest.mark.parametrize(
    "value,parser", [(v, cp.List(p)) for (v, p) in FAILURE_PRIMITIVES]
)
def test_failed_list_parsers(value: t.Any, parser: cp.Parser[t.Any]):
    with pytest.raises(Exception):
        parser(value)


@pytest.mark.parametrize(
    "value,parser,expected",
    [(v + "," + v, cp.List(p), [e, e]) for (v, p, e) in SUCCESS_PRIMITIVES],
)
def test_successfull_two_item_list_parsers(
    value: t.Any, parser: cp.Parser[t.Any], expected: t.Any
):
    assert parser(value) == expected


@pytest.mark.parametrize(
    "value,parser,expected",
    [(v, cp.Tuple(p, num=1), (e,)) for (v, p, e) in SUCCESS_PRIMITIVES],
)
def test_successfull_tuple_parsers(
    value: t.Any, parser: cp.Parser[t.Any], expected: t.Any
):
    assert parser(value) == expected


@pytest.mark.parametrize(
    "value,parser",
    [(v, cp.Tuple(p, num=1)) for (v, p) in FAILURE_PRIMITIVES],
)
def test_failed_tuple_parsers(value: t.Any, parser: cp.Parser[t.Any]):
    with pytest.raises(Exception):
        parser(value)


@pytest.mark.parametrize(
    "value,parser,expected",
    [(v + "," + v, cp.Tuple(p, p, num=2), (e, e)) for (v, p, e) in SUCCESS_PRIMITIVES],
)
def test_successfull_two_item_tuple_parsers(
    value: t.Any, parser: cp.Parser[t.Any], expected: t.Any
):
    assert parser(value) == expected


@pytest.mark.parametrize(
    "_type,expected",
    [
        (Positional[int], cp.Int()),
        (int, cp.Int()),
        (float, cp.Float()),
        (bool, cp.Bool()),
        (str, cp.Str()),
        (datetime, cp.DateTime()),
        (timedelta, cp.TimeDelta()),
        (Path, cp.Path()),
        (int | bool, cp.Union(cp.Int(), cp.Bool())),
        (int | bool | str, cp.Int() | cp.Bool() | cp.Str()),
        (t.Literal[1, "foo"], cp.Literal("1", "foo")),
        (Color, cp.Enum(Color)),
        (int | None, cp.Union(cp.Int(), cp.NoneParser())),
        (t.Optional[int], cp.Union(cp.Int(), cp.NoneParser())),
    ],
)
def test_parser_from_type(_type: t.Any, expected: cp.Parser[t.Any]):
    assert cp.from_type(_type) == expected


@pytest.mark.parametrize(
    "parser,expected",
    [
        (cp.Int(), "integer"),
        (cp.Float(), "float"),
        (cp.Bool(), "{yes|no}"),
        (cp.Str(), "text"),
        (cp.DateTime(), "datetime"),
        (cp.TimeDelta(), "timedelta"),
        (cp.Path(), "path"),
        (cp.Union(cp.Int(), cp.Bool()), "(integer|yes|no)"),
        (cp.Int() | cp.Bool() | cp.Str(), "(integer|yes|no|text)"),
        (cp.Literal("1", "foo"), "{1|foo}"),
        (cp.Enum(Color), "{red|blue}"),
    ],
)
def test_parser_str(parser: cp.Parser[t.Any], expected: cp.Parser[t.Any]):
    assert str(parser) == expected


def test_parse_tuple_list():
    parser = cp.List(cp.Tuple(cp.Str(), cp.Int(), num=2))
    assert parser(["a,1", "b,2", "c,3"]) == [
        ("a", 1),
        ("b", 2),
        ("c", 3),
    ]


@pytest.mark.parametrize(
    "parser,input,expected",
    [
        (cp.DateTime(tz=UTC), "2025-5-15", datetime(2025, 5, 15, 0, 0, 0, tzinfo=UTC)),
        (
            cp.DateTime(tz=UTC),
            "2025-05-15T00:00:00Z",
            datetime(2025, 5, 15, 0, 0, 0, tzinfo=UTC),
        ),
        (
            cp.DateTime(tz=UTC),
            "2025-05-15T00:00:00-01:00",
            datetime(2025, 5, 15, 1, 0, 0, tzinfo=UTC),
        ),
    ],
)
def test_date_with_tz(parser: cp.DateTime, input: str, expected: datetime):
    assert parser(input) == expected



from pathlib import Path

from typing_extensions import override

import clypi.parsers as cp
from clypi import Command, Positional, arg


class Lint(Command):
    files: Positional[tuple[Path, ...]]
    verbose: bool = arg(inherited=True)  # Comes from MyCli but I want to use it too

    @override
    async def run(self):
        print(f"Linting {self.files=} and {self.verbose=}")


class MyCli(Command):
    """
    my-cli is a very nifty demo CLI tool
    """

    subcommand: Lint | None = None
    threads: int = arg(
        default=4,
        # Built-in parsers for useful validations
        parser=cp.Int(min=1, max=10),
    )
    verbose: bool = arg(
        False,
        help="Whether to show extra logs",
        prompt="Do you want to see extra logs?",
        short="v",  # User can pass in --verbose or -v
    )

    @override
    async def run(self):
        print(f"Running the main command with {self.verbose}")


if __name__ == "__main__":
    cli: MyCli = MyCli.parse()
    cli.start()



import typing as t
from types import NoneType

import pytest

from clypi._type_util import (
    is_list,
    is_literal,
    is_optional,
    is_tuple,
    is_union,
    literal_inner,
    tuple_inner,
    union_inner,
)


@pytest.mark.parametrize(
    "_type",
    [
        list[str],
        list[tuple[int]],
        t.Annotated[list[str], 1],
    ],
)
def test_is_list(_type: t.Any):
    assert is_list(_type)


@pytest.mark.parametrize(
    "_type",
    [
        tuple[str],
        tuple[list[int]],
        t.Annotated[tuple[str], 1],
    ],
)
def test_is_tuple(_type: t.Any):
    assert is_tuple(_type)


@pytest.mark.parametrize(
    "_type",
    [
        str | int,
        str | int | bool,
        t.Union[int, bool],
        t.Annotated[str | int, 1],
        t.Annotated[t.Union[str | int], 1],
    ],
)
def test_is_union(_type: t.Any):
    assert is_union(_type)


@pytest.mark.parametrize(
    "_type",
    [
        t.Literal["a"],
        t.Literal["a", 1],
        t.Annotated[t.Literal["a", 1], 1],
    ],
)
def test_is_literal(_type: t.Any):
    assert is_literal(_type)


@pytest.mark.parametrize(
    "_type",
    [bool | None, None | bool, t.Optional[bool]],
)
def test_is_optional(_type: t.Any):
    assert is_optional(_type)


@pytest.mark.parametrize(
    "_type,inner_t,num",
    [
        (tuple[str], [str], 1),
        (tuple[str, int], [str, int], 2),
        (tuple[str, ...], [str], None),
    ],
)
def test_tuple_inner(_type: t.Any, inner_t: t.Any, num: int):
    assert tuple_inner(_type) == (inner_t, num)


@pytest.mark.parametrize(
    "_type,inner_t",
    [
        (str | int, [str, int]),
        (str | int | None, [str, int, NoneType]),
        (t.Union[str, int, None], [str, int, NoneType]),
    ],
)
def test_union_inner(_type: t.Any, inner_t: t.Any):
    assert union_inner(_type) == inner_t


@pytest.mark.parametrize(
    "_type,inner",
    [
        (t.Literal["a"], ["a"]),
        (t.Literal["a", 1], ["a", 1]),
    ],
)
def test_literal_inner(_type: t.Any, inner: t.Any):
    assert literal_inner(_type) == inner



from typing_extensions import override

from clypi import Command, arg, cprint, style


class Main(Command):
    """An example of how enabling negative flags looks like"""

    verbose: bool = arg(
        True,
        short="v",
        negative="quiet",
        help="Whether to show more output",
    )

    @override
    async def run(self):
        cprint(f"Verbose: {self.verbose}", fg="blue")
        print(
            style("Try using ", fg="cyan")
            + style("--quiet", fg="yellow", bold=True)
            + style(" or ", fg="cyan")
            + style("--help", fg="yellow", bold=True)
        )


if __name__ == "__main__":
    main: Main = Main.parse()
    main.start()



import typing as t
from dataclasses import dataclass, field

from clypi._data.dunders import ALL_DUNDERS
from clypi._prompts import MAX_ATTEMPTS, prompt
from clypi._util import UNSET, Unset
from clypi.parsers import Parser

T = t.TypeVar("T")


def gen_impl(__f: str) -> t.Callable[..., t.Any]:
    def _impl(self: "DeferredValue[t.Any]", *args: t.Any, **kwargs: t.Any) -> t.Any:
        return getattr(self.__get__(None), __f)(*args, **kwargs)

    return _impl


@dataclass
class DeferredValue(t.Generic[T]):
    parser: Parser[T]
    prompt: str
    default: T | Unset = UNSET
    default_factory: t.Callable[[], T] | Unset = UNSET
    hide_input: bool = False
    max_attempts: int = MAX_ATTEMPTS

    _value: T | Unset = field(init=False, default=UNSET)

    def __set_name__(self, owner: t.Any, name: str):
        self.__name__ = name

    def __get__(self, obj: t.Any, objtype: t.Any = None) -> T:
        if self._value is UNSET:
            self._value = prompt(
                self.prompt,
                default=self.default,
                default_factory=self.default_factory,
                hide_input=self.hide_input,
                max_attempts=self.max_attempts,
                parser=self.parser,
            )
        return self._value

    # Autogen all dunder methods to trigger __get__
    # NOTE: I hate having to do this but I did not find how to trigger
    # the evaluation of a descriptor when a dunder method is called on it
    for dunder in ALL_DUNDERS:
        locals()[dunder] = gen_impl(dunder)



from __future__ import annotations

import typing as t
from enum import Enum
from getpass import getpass

import clypi
from clypi import parsers
from clypi._configuration import get_config
from clypi._exceptions import AbortException, MaxAttemptsException
from clypi._util import UNSET, Unset

MAX_ATTEMPTS: int = 20


def _error(msg: str):
    clypi.cprint(msg, fg="red")


def _input(prompt: str, hide_input: bool = False) -> str:
    """
    Prompts the user for a value or uses the default and returns the
    value and if we're using the default
    """
    fun = getpass if hide_input else input
    styled_prompt = get_config().theme.prompts(prompt)
    return fun(styled_prompt)


def _display_default(default: t.Any) -> str:
    if isinstance(default, bool):
        return "Y/n" if default else "y/N"
    if isinstance(default, Enum):
        return default.name.lower()
    return f"{default}"


def _build_prompt(text: str, default: t.Any | Unset) -> str:
    prompt = text
    if default is not UNSET:
        prompt += f" [{_display_default(default)}]"
    prompt += ": "
    return prompt


def confirm(
    text: str,
    *,
    default: bool | Unset = UNSET,
    default_factory: t.Callable[[], bool] | Unset = UNSET,
    max_attempts: int = MAX_ATTEMPTS,
    abort: bool = False,
) -> bool:
    """
    Prompt the user for a yes/no value

    :param text: The prompt text.
    :param default: The default value.
    :param max_attempts: The maximum number of attempts to get a valid value.
    :return: The parsed value.
    """
    parsed_inp = prompt(
        text=text,
        default=default,
        default_factory=default_factory,
        max_attempts=max_attempts,
        parser=parsers.from_type(bool),
    )
    if abort and not parsed_inp:
        raise AbortException()
    return parsed_inp


T = t.TypeVar("T")


@t.overload
def prompt(
    text: str,
    *,
    default: str | Unset = UNSET,
    default_factory: t.Callable[[], str] | Unset = UNSET,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
) -> str: ...


@t.overload
def prompt(
    text: str,
    *,
    parser: parsers.Parser[T] | type[T],
    default: T | Unset = UNSET,
    default_factory: t.Callable[[], T] | Unset = UNSET,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
) -> T: ...


def prompt(
    text: str,
    *,
    parser: parsers.Parser[T] | type[T] | type[str] = str,
    default: T | Unset = UNSET,
    default_factory: t.Callable[[], T] | Unset = UNSET,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
) -> T:
    """
    Prompt the user for a value.

    :param text: The prompt text.
    :param default: The default value.
    :param parser: The parser function parse the input with.
    :param max_attempts: The maximum number of attempts to get a valid value.
    :return: The parsed value.
    """
    if default_factory is not UNSET:
        default = default_factory()

    # Build the prompt
    prompt = _build_prompt(text, default)

    # Loop until we get a valid value
    for _ in range(max_attempts):
        inp = _input(prompt, hide_input=hide_input)
        if not inp and default is UNSET:
            _error("A value is required.")
            continue

        # User answered the prompt -- Parse
        try:
            if t.TYPE_CHECKING:
                parser = t.cast(parsers.Parser[T], parser)

            # If no input, use the default without parsing
            if not inp and default is not UNSET:
                parsed_inp = default

            # Otherwise try parsing the string
            else:
                parsed_inp = parser(inp)
        except parsers.CATCH_ERRORS as e:
            _error(f"Unable to parse {inp!r}, please provide a valid value.\n  ↳ {e}")
            continue

        return parsed_inp

    raise MaxAttemptsException(
        f"Failed to get a valid value after {max_attempts} attempts."
    )



import typing as t

from clypi._colors import ColorType, Styler
from clypi._util import get_term_width, visible_width


def separator(
    separator: str = "━",
    width: t.Literal["max"] | int = "max",
    title: str | None = None,
    color: ColorType | None = None,
) -> str:
    if width == "max":
        width = get_term_width()

    c = Styler(fg=color)

    if not title:
        return c(separator * width)

    num_chars = width - 2 - visible_width(title)
    left_chars = num_chars // 2
    right_chars = num_chars - left_chars
    return c(separator * left_chars + " " + title + " " + separator * right_chars)



from dataclasses import dataclass, field

from clypi._cli.arg_config import Nargs


@dataclass
class CurrentCtx:
    name: str = ""
    nargs: Nargs = 0
    max_nargs: Nargs = 0

    _collected: list[str] = field(init=False, default_factory=list)

    def has_more(self) -> bool:
        if isinstance(self.nargs, float | int):
            return self.nargs > 0
        return True

    def needs_more(self) -> bool:
        if isinstance(self.nargs, float | int):
            return self.nargs > 0
        return False

    def collect(self, item: str) -> None:
        if isinstance(self.nargs, float | int):
            self.nargs -= 1

        self._collected.append(item)

    @property
    def collected(self) -> str | list[str]:
        if self.max_nargs == 1:
            return self._collected[0]
        return self._collected

    def __bool__(self) -> bool:
        return bool(self.name)



import typing as t

import pytest

from clypi._cli.arg_config import Nargs, _get_nargs  # type: ignore


@pytest.mark.parametrize(
    "_type,expected",
    [
        (bool, 0),
        (t.Literal["a"], 1),
        (t.Optional[bool], 1),
        (list[bool], "*"),
        (list[bool] | None, "*"),
        (list[bool] | None, "*"),
        (bool | int, 1),
    ],
)
def test_get_nargs(_type: t.Any, expected: Nargs):
    assert _get_nargs(_type) == expected



import pytest
from typing_extensions import override

from clypi import Command

# A counter to check the order of events and their results
_counter: list[Exception | int] = []

# An exception we will raise and catch
_EXC: Exception = Exception("foo")


class ExampleSubCommand(Command):
    """Some sample docs"""

    should_raise: bool = False

    @override
    async def pre_run_hook(self) -> None:
        _counter.append(1)

    @override
    async def post_run_hook(self, exception: Exception | None) -> None:
        _counter.append(exception or 3)

    @override
    async def run(self):
        _counter.append(2)

        if self.should_raise:
            raise _EXC


def test_cli_hooks_run_in_order():
    global _counter
    _counter = []

    ExampleSubCommand().parse([]).start()

    # Pre-run, run, post-run no exception
    assert _counter == [1, 2, 3]


def test_cli_hooks_catch_exception():
    global _counter
    _counter = []

    with pytest.raises(Exception):
        ExampleSubCommand().parse(["--should-raise"]).start()

    # Pre-run, run, post-run with catch
    assert _counter == [1, 2, _EXC]



from clypi import separator


def test_default_separator():
    assert separator(width=20) == "━" * 20


def test_separator_with_title():
    expected = "━" * 19 + " Test " + "━" * 20
    assert separator(width=45, title="Test") == expected



import asyncio
import sys

from clypi import Spin, Spinner, cprint, spinner


async def all_spinners():
    cprint(
        "Displaying all spinner animations." + "\n ↳ Press ctrl+c to skip all examples",
        fg="blue",
        bold=True,
    )

    for i, anim in enumerate(Spin, 1):
        async with Spinner(
            f"{anim.human_name()} spinning animation [{i}/{len(Spin)}]",
            animation=anim,
        ):
            await asyncio.sleep(1.2)


async def subprocess():
    # Example with subprocess
    title = "Example with subprocess"
    async with Spinner(title) as s:
        # Fist subprocess
        proc = await asyncio.create_subprocess_shell(
            "for i in $(seq 1 10); do date && sleep 0.2; done;",
            stdout=asyncio.subprocess.PIPE,
        )

        # Second subprocess
        proc2 = await asyncio.create_subprocess_shell(
            "for i in $(seq 1 20); do echo $RANDOM && sleep 0.1; done;",
            stdout=asyncio.subprocess.PIPE,
        )

        coros = (
            s.pipe(proc.stdout, color="red"),
            s.pipe(proc2.stdout, prefix="(rand)"),
        )
        await asyncio.gather(*coros)


@spinner("Example that captures stdout/stderr", capture=True)
async def captured_with_decorator():
    # Example with subprocess
    for i in range(10):
        if i % 2 == 0:
            cprint("Stdout output", fg="blue")
        else:
            cprint("Stderr output", fg="red", file=sys.stderr)
        await asyncio.sleep(0.3)


async def main():
    # Run all of the spinner animations
    try:
        await all_spinners()
    except asyncio.CancelledError:
        pass

    # Display a subprocess example
    print()
    try:
        await subprocess()
    except asyncio.CancelledError:
        pass

    # Show an example decorator usage with stdout capture
    print()
    await captured_with_decorator()


if __name__ == "__main__":
    asyncio.run(main())



from typing import overload

from clypi._components.wraps import wrap
from clypi._util import get_term_width, visible_width


def _safe_get(ls: list[str], idx: int) -> str:
    if idx >= len(ls):
        return ""
    return ls[idx]


@overload
def stack(
    *blocks: list[str],
    max_width: int | None = None,
    padding: int = 2,
    lines: bool,
) -> list[str]: ...


@overload
def stack(
    *blocks: list[str],
    max_width: int | None = None,
    padding: int = 2,
) -> str: ...


def stack(
    *blocks: list[str],
    max_width: int | None = None,
    padding: int = 2,
    lines: bool = False,
) -> str | list[str]:
    max_width = max_width or get_term_width()
    padding_str = " " * padding

    new_lines: list[str] = []
    height = max(len(b) for b in blocks)
    width_per_block = [max(visible_width(line) for line in block) for block in blocks]

    # Process line until all blocks are done
    for idx in range(height):
        more = False
        tmp: list[str] = []

        # Add the line from each block into combined line
        for block, block_width in zip(blocks, width_per_block):
            # If there was a line, next iter will happen
            block_line = _safe_get(block, idx)
            if block_line:
                more |= True

            # How much do we need to reach the actual visible length
            actual_width = (block_width - visible_width(block_line)) + len(block_line)

            # Align and append line
            tmp.append(block_line.ljust(actual_width))

        # Check if combined line would overflow and wrap if needed
        combined_line = padding_str.join(tmp).rstrip()
        if visible_width(combined_line) <= max_width:
            new_lines.append(combined_line)
        else:
            # We need to wrap the last block and the remainder needs to be aligned
            # with the start of the second block
            width_without_last = visible_width(padding_str.join(tmp[:-1]) + padding_str)
            max_last_width = max_width - width_without_last
            wrapped_last = wrap(tmp[-1], max_last_width)

            # Add the combined line
            combined_line = padding_str.join(tmp[:-1] + [wrapped_last[0]]).rstrip()
            new_lines.append(combined_line)

            # Add the remainder aligned to the start of second block
            padding_left = " " * width_without_last
            for remaining_line in wrapped_last[1:]:
                new_lines.append(padding_left + remaining_line)

        # Exit if no more lines in any iter
        if not more:
            break

    return new_lines if lines else "\n".join(new_lines)



import asyncio
import typing as t
from enum import Enum
from pathlib import Path

from typing_extensions import override

import clypi.parsers as cp
from clypi import Command, Positional, Spinner, arg, boxed, cprint, style


# ---- START DEMO UTILS ----
def debug(command: Command) -> None:
    """
    Just a utility function to display the commands being passed in a somewhat
    nice way
    """
    box = boxed(style(command, bold=True), title="Debug", color="magenta")
    print(box, end="\n\n")


# ---- END DEMO UTILS ----


class Env(Enum):
    QA = 1
    PROD = 2


class RunParallel(Command):
    """
    Runs all of the files in parallel
    """

    files: Positional[list[str]]
    exceptions_with_reasons: Path | None = arg(None, parser=cp.Path(exists=True))
    env: Env = arg(inherited=True)

    @override
    async def run(self):
        debug(self)
        cprint(f"{self.env.name} - Running all files", fg="blue", bold=True)

        async with Spinner(f"Running {', '.join(self.files)} in parallel"):
            await asyncio.sleep(2)

        async with Spinner(f"Linting {', '.join(self.files)} in parallel"):
            await asyncio.sleep(2)

        cprint("\nDone!", fg="green", bold=True)


class RunSerial(Command):
    """
    Runs all of the files one by one
    """

    files: Positional[list[Path]] = arg(parser=cp.List(cp.Path(exists=True)))
    env: Env = arg(inherited=True)

    @override
    async def run(self):
        debug(self)
        cprint(f"{self.env.name} - Running all files", fg="blue", bold=True)
        for f in self.files:
            async with Spinner(f"Running {f.as_posix()} in parallel"):
                await asyncio.sleep(2)
        cprint("\nDone!", fg="green", bold=True)


class Run(Command):
    """
    Allows running files with different options
    """

    subcommand: RunParallel | RunSerial
    quiet: bool = arg(
        False,
        short="q",
        help="If the runner should omit all stdout messages",
        group="Global",
    )
    env: Env = arg(Env.PROD, help="The environment to run in")
    format: t.Literal["json", "pretty"] = arg(
        "pretty", help="The format with which to display results"
    )


class Lint(Command):
    """
    Lints all of the files in a given directory using the latest
    termuff rules.
    """

    files: Positional[list[str]] = arg(help="The list of files to lint")
    quiet: bool = arg(
        False,
        short="q",
        help="If the linter should omit all stdout messages",
    )
    timeout: int = arg(help="Disable the termuff cache")
    index: str = arg(
        "http://pypi.org",
        help="The index to download termuff from",
        prompt="What index do you want to download termuff from?",
    )

    @override
    async def run(self) -> None:
        debug(self)
        async with Spinner(f"Linting {', '.join(self.files)}"):
            await asyncio.sleep(self.timeout)
        cprint("\nDone!", fg="green", bold=True)


class Main(Command):
    """
    Termuff is a powerful command line interface to lint and
    run arbitrary files.
    """

    subcommand: Run | Lint | None = None
    verbose: bool = arg(False, short="v", help="Whether to show more output")

    @override
    @classmethod
    def prog(cls):
        return "termuff"

    @override
    @classmethod
    def epilog(cls):
        return "Learn more at http://termuff.org"


if __name__ == "__main__":
    main: Main = Main.parse()
    main.start()
